import numpy as np
import networkx as nx
from sklearn.neighbors import NearestNeighbors
from skimage import morphology
import scipy
import matplotlib.pyplot as plt
import math
from scipy.interpolate import RegularGridInterpolator 


from scipy.spatial import distance


def curveorder(x,y):
    """
    Sort pixels that make up the filament spine by the order in which they appear along the curve
    Code taken from http://stackoverflow.com/questions/37742358/sorting-points-to-form-a-continuous-line

    Parameters:

    x: numpy.ndarray
       A 1d array specifying the x coordinates of the pixels defining your filament spine
    y: numpy.ndarray
       A 1d array specifying the y coordinates of the pixels defining your filament spine

    Returns:
    xx, yy: the sorted x and y arrays given above

    """

    # make the point list (N x 2)
    pts=np.vstack((x,y)).T

    # initiate the NN2 and the network graph
    clf = NearestNeighbors(2).fit(pts)
    G = clf.kneighbors_graph()
    T = nx.from_scipy_sparse_matrix(G)

    # candidate paths based on the network graph
    paths = [list(nx.dfs_preorder_nodes(T, i)) for i in range(len(pts))]
    ## find the path with the lowest cost (distance) among the candidates
    minidx = np.argmin([np.sum(np.diagonal(distance.cdist(pts[path], pts[path]), offset = 1)) for path in paths])
    opt_order = paths[minidx]
    ## permute the head and the tail to find the correct order
    ### permutation
    opt_order_swaphead = list(opt_order)
    opt_order_swaphead[0], opt_order_swaphead[1] = opt_order_swaphead[1], opt_order_swaphead[0]
    opt_order_swaptail = list(opt_order)
    opt_order_swaptail[-1], opt_order_swaptail[-2] = opt_order_swaptail[-2], opt_order_swaptail[-1]
    ### find the correct order among the original and the two permuted
    paths_opt = [opt_order, opt_order_swaphead, opt_order_swaptail]
    minidx_opt = np.argmin([np.sum(np.diagonal(distance.cdist(pts[path], pts[path]), offset = 1)) for path in paths_opt])
    opt_order_final = paths_opt[minidx_opt]

    # return the ordered coordinates
    xx = x[opt_order_final]
    yy = y[opt_order_final]
    ## make it always go in the increasing y direction
    if yy[-1] < yy[0]:
        yy = yy[::-1]
        xx = xx[::-1]

    return(xx,yy)


def profile_builder(radobj, point, derivative, shift = True, fold = False):
    '''
    Build the profile using array manipulation, instead of looping.

    Parameters:

    radobj:
    The object containing the image, the mask, and the axis for plotting.

    point: tuple-like
    The x and y pixel coordinates (corresponding to the 1st and the 0th axes) of
    the point at the center.

    derivative: tuple-like
    Thee x and y components of the derivative of the spline at `point`.  Used to
    derive the profile cut.

    shift: boolean
    Indicates whether to shift the profile to center at the peak value.

    fold: boolean
    Indicates whether to fold around the central pixel, so that the final profile
    will be a "half profile" with the peak near/at the center (depending on
    whether it's shifted).


    Returns:

    final_dist: 1D numpy.ndarray
    The distance array.

    image_line: 1D numpy.ndarray
    The value array.

    '''
    # Read the image and the mask
    image, mask = radobj.image, radobj.mask
    # Read the plotting axis
    axis = radobj.ax

    # Read the point and double check whether it's inside the mask.
    x0, y0 = point
    if (not mask[int(round(y0)), int(round(x0))]):
        raise ValueError("The point is not in the mask.")

    # Create the grid to calculate where the profile cut crosses edges of the
    # pixels.
    shapex, shapey = image.shape[1], image.shape[0]
    edgex, edgey = np.arange(.5, shapex-.5, 1.), np.arange(.5, shapey-.5, 1.)

    # Extreme cases when the derivative is (1, 0) or (0, 1)
    if (derivative[0] == 0) or (derivative[1] == 0):
        if (derivative[0] == 0) and (derivative[1] == 0):
            raise ValueError("Both components of the derivative are zero; unable to derive a tangent.")
        elif (derivative[0] == 0):
            y_edgex = []
            edgex = []
            x_edgey = np.ones(len(edgey))*x0
        elif (derivative[1] == 0):
            y_edgex = np.ones(len(edgex))*y0
            x_edgey = []
            edgey = []

    ## The regular cases go here: calculate the crossing points of the cut and the grid.
    else:
        slope = -1./(derivative[1]/derivative[0])
        y_edgex = slope*(edgex - x0) + y0
        x_edgey = (edgey - y0)/slope + x0

        ### Mask out points outside the image.
        pts_maskx = ((np.round(x_edgey) >= 0.) & (np.round(x_edgey) < shapex))
        pts_masky = ((np.round(y_edgex) >= 0.) & (np.round(y_edgex) < shapey))

        edgex, edgey = edgex[pts_masky], edgey[pts_maskx]
        y_edgex, x_edgey = y_edgex[pts_masky], x_edgey[pts_maskx]


    # Sort the points to find the center of each segment inside a single pixel.
    ## This also deals with when the cut crosses at the 4-corner point(s).
    ## The sorting is done by sorting the x coordinates
    stack = sorted(list(set(zip(np.concatenate([edgex, x_edgey]),\
                       np.concatenate([y_edgex, edgey])))))
    centers =  stack[:-1]+.5*np.diff(stack, axis = 0)

    ## extract the values from the image and the original mask
    #setup interpolation 
    xgrid=np.arange(0.5,radobj.image.shape[1]+0.5,1.0)
    ygrid=np.arange(0.5,radobj.image.shape[0]+0.5,1.0)
    interpolator = RegularGridInterpolator((xgrid,ygrid),radobj.image.T,bounds_error=False,fill_value=None)
    
    image_line=interpolator(centers)
    #image_line = image[np.round(centers[:, 1]).astype(int), np.round(centers[:, 0]).astype(int)]
    
    mask_line = mask[np.round(centers[:, 1]).astype(int), np.round(centers[:, 0]).astype(int)]
    #### select the part of the mask that includes the original point
    mask_p0 = (np.round(centers[:, 0]).astype(int) == int(round(x0)))&\
              (np.round(centers[:, 1]).astype(int) == int(round(y0)))
    mask_line = (morphology.label(mask_line) == morphology.label(mask_line)[mask_p0])

    # Extract the profile from the image.
    ## for the points within the original mask; to find the peak
    if derivative[1] < 0.:
        image_line0 = image_line[mask_line][::-1]
        centers = centers[::-1]
        mask_line = mask_line[::-1]
        mask_p0 = mask_p0[::-1]
    else:
        image_line0 = image_line[mask_line]
    ## for the entire map
    if derivative[1] < 0.:
        image_line = image_line[::-1]
    else:
        image_line = image_line

    # Plot.
    peak_finder = centers[mask_line]
    ## find the end points of the cuts (within the original mask)
    start, end = peak_finder[0], peak_finder[-1]
    ## find the peak here
    xpeak, ypeak = peak_finder[image_line0 == np.nanmax(image_line0)][0]
    ## the peak mask is used to determine where to unfold when shift = True
    mask_peak = (np.round(centers[:, 0]).astype(int) == int(round(xpeak)))&\
                (np.round(centers[:, 1]).astype(int) == int(round(ypeak)))
    ## plot the cut
    axis.plot([start[0], end[0]], [start[1], end[1]], 'r-', linewidth = 1.,alpha=1)

    # Shift.
    if shift:
        final_dist = np.hypot(centers[:, 0]-xpeak, centers[:, 1]-ypeak)
        # unfold
        pos0 = np.where(mask_peak)[0][0]
        final_dist[:pos0] = -final_dist[:pos0]
    else:
        final_dist = np.hypot(centers[:, 0]-x0, centers[:, 1]-y0)
        # unfold
        pos0 = np.where(mask_p0)[0][0]
        final_dist[:pos0] = -final_dist[:pos0]


    # Fold
    if fold:
            final_dist = abs(final_dist)

    return final_dist, image_line, (xpeak, ypeak), (start, end)



import numpy as np
import numbers
import matplotlib.colors as colors
import matplotlib.pyplot as plt
from astropy.wcs import WCS
from . import styles


def plotCuts(radobj, ax):

    if hasattr(radobj, 'dictionary_cuts'):
        dictionary_cuts = radobj.dictionary_cuts.copy()
    else:
        raise ValueError('Please run build_profile before plotting.')

    # plot the peaks
    if dictionary_cuts['plot_peaks'] is not None:
        toPlot = np.asarray(dictionary_cuts['plot_peaks'])

        ax.plot(toPlot[:, 0].astype(int), toPlot[:, 1].astype(int), 'b.',
                markersize = 12.,
                alpha=0.75,
                zorder = 999, markeredgecolor='white',markeredgewidth=0.5)

    # plot the cuts
    if dictionary_cuts['plot_cuts'] is not None:
        toPlot = dictionary_cuts['plot_cuts']

        [ax.plot(np.asarray(cut)[:, 0], np.asarray(cut)[:, 1], 'r-', linewidth = 1.)\
         for cut in toPlot]


    return ax


class RadFilPlotter(object):

    '''
    A class to plot the results in radfil objects.
    '''

    def __init__(self, radobj):
        self.radobj = radobj


    def plotCuts(self, ax, savefig = False):

        ## prepare
        vmin, vmax = np.min(self.radobj.image[self.radobj.mask]), np.nanpercentile(self.radobj.image[self.radobj.mask], 98.)
        xmin, xmax = np.where(self.radobj.mask)[1].min(), np.where(self.radobj.mask)[1].max()
        ymin, ymax = np.where(self.radobj.mask)[0].min(), np.where(self.radobj.mask)[0].max()


        if self.radobj.cutting:


            ## plotting
            ax.imshow(self.radobj.image,
                      origin='lower',
                      cmap='gray',
                      interpolation='none',
                      norm = colors.Normalize(vmin = vmin, vmax =  vmax))
            ax.contourf(self.radobj.mask,
                        levels = [0., .5],
                        colors = 'w')
            ax.plot(self.radobj.xspline, self.radobj.yspline, 'r', label='fit', lw=3, alpha=1.0)
            ax.set_xlim(max(0., xmin-.1*(xmax-xmin)), min(self.radobj.mask.shape[1]-.5, xmax+.1*(xmax-xmin)))
            ax.set_ylim(max(0., ymin-.1*(ymax-ymin)), min(self.radobj.mask.shape[0]-.5, ymax+.1*(ymax-ymin)))

        else:

            ## plotting
            ax.imshow(self.radobj.image,
                      origin='lower',
                      cmap='gray',
                      interpolation='none',
                      norm = colors.Normalize(vmin = vmin, vmax =  vmax))
            ax.contourf(self.radobj.mask,
                        levels = [0., .5],
                        colors = 'w')
            ax.plot(line.xy[0], line.xy[1], 'r', label='fit', lw=2, alpha=0.5)
            ax.set_xlim(max(0., xmin-.1*(xmax-xmin)), min(self.radobj.mask.shape[1]-.5, xmax+.1*(xmax-xmin)))
            ax.set_ylim(max(0., ymin-.1*(ymax-ymin)), min(self.radobj.mask.shape[0]-.5, ymax+.1*(ymax-ymin)))


        plotCuts(self.radobj, ax)


    def plotFits(self, ax, plotFeature):

        if isinstance(plotFeature, str):
            if plotFeature.lower() == 'model':

                if self.radobj.bgdist is not None:
                    xplot = self.radobj.xall
                    yplot = self.radobj.yall - self.radobj.bgfit(xplot)
                    xlim=np.max(self.radobj.bgdist*1.5)


                else:
                    xplot=self.radobj.xall
                    yplot=self.radobj.yall
                    xlim=np.max(np.absolute(self.radobj.fitdist))*1.5


                ## Plot model
                #Adjust axis limit based on percentiles of data
                #axis.set_xlim(np.min(self.radobj.xall), np.max(self.radobj.xall))
                #xlim=np.max(np.absolute([np.nanpercentile(self.radobj.xall[np.isfinite(self.radobj.yall)],1),np.nanpercentile(self.radobj.xall[np.isfinite(self.radobj.yall)],99)]))

                if not self.radobj.fold:
                    ax.set_xlim(-xlim,+xlim)
                else:
                    ax.set_xlim(0., +xlim)
                ax.set_ylim(np.nanpercentile(yplot,0)-np.abs(0.5*np.nanpercentile(yplot,0)),np.nanpercentile(yplot,99.9)+np.abs(0.25*np.nanpercentile(yplot,99.9)))


                ax.plot(xplot, yplot, 'k.', markersize = 1., alpha=styles.get_scatter_alpha(len(self.radobj.xall))) 
                if self.radobj.binning:
                    if self.radobj.bgdist is not None:
                        plotbinx, plotbiny = np.ravel(list(zip(self.radobj.bins[:-1], self.radobj.bins[1:]))), np.ravel(list(zip(self.radobj.mastery-self.radobj.bgfit(self.radobj.masterx), self.radobj.mastery-self.radobj.bgfit(self.radobj.masterx))))
                    else:
                        plotbinx, plotbiny = np.ravel(list(zip(self.radobj.bins[:-1], self.radobj.bins[1:]))), np.ravel(list(zip(self.radobj.mastery, self.radobj.mastery)))
                    ax.plot(plotbinx, plotbiny,
                              'r-')

                # Plot the range
                if self.radobj.fitdist is not None:
                    ## symmetric fitting range
                    if isinstance(self.radobj.fitdist, numbers.Number):
                        ax.fill_between([-self.radobj.fitdist, self.radobj.fitdist], *ax.get_ylim(),
                                          facecolor = (0., 0., 1., .05),
                                          edgecolor = 'b',
                                          linestyle = '--',
                                          linewidth = 1.)
                    ## asymmetric fitting range
                    elif np.asarray(self.radobj.fitdist).shape == (2,):
                        plot_fitdist = self.radobj.fitdist.copy()
                        plot_fitdist[~np.isfinite(plot_fitdist)] = np.asarray(ax.get_xlim())[~np.isfinite(plot_fitdist)]
                        ax.fill_between(plot_fitdist, *ax.get_ylim(),
                                          facecolor = (0., 0., 1., .05),
                                          edgecolor = 'b',
                                          linestyle = '--',
                                          linewidth = 1.)
                ## no fitting range; all data are used
                else:
                    ax.fill_between(ax.get_xlim(), *ax.get_ylim(),
                                      facecolor = (0., 0., 1., .05),
                                      edgecolor = 'b',
                                      linestyle = '--',
                                      linewidth = 1.)

                # Plot the predicted curve
                ax.plot(np.linspace(ax.get_xlim()[0],ax.get_xlim()[1],500), self.radobj.profilefit(np.linspace(ax.get_xlim()[0],ax.get_xlim()[1],500)), 'b-', lw = 3., alpha = .6)


                ax.text(0.03, 0.95,"{}={:.2E}\n{}={:.2f}\n{}={:.2f}".format(self.radobj.profilefit.param_names[0],self.radobj.profilefit.parameters[0],self.radobj.profilefit.param_names[1],self.radobj.profilefit.parameters[1],self.radobj.profilefit.param_names[2],self.radobj.profilefit.parameters[2]),ha='left',va='top', fontweight='bold',fontsize=20,transform=ax.transAxes)#,bbox={'facecolor':'white', 'edgecolor':'none', 'alpha':1.0, 'pad':1})
                ax.text(0.97, 0.95,"{}\nFit".format(self.radobj.fitfunc.capitalize()), ha='right',va='top', color='blue',fontweight='bold', fontsize=20, transform=ax.transAxes)#,bbox={'facecolor':'white', 'edgecolor':'none', 'alpha':1.0, 'pad':1})
                #ax.tick_params(labelsize=14)


            elif plotFeature.lower() == 'bg':

                if self.radobj.bgdist is None:
                    raise ValueError('No bgfit in the radfil object. Rerun fit_profile.')

                #xlim=np.max(np.absolute([np.nanpercentile(self.radobj.xall[np.isfinite(self.radobj.yall)],1),np.nanpercentile(self.radobj.xall[np.isfinite(self.radobj.yall)],99)]))
                xlim=np.max(self.radobj.bgdist*1.5)

                if not self.radobj.fold:
                    ax.set_xlim(-xlim,+xlim)
                else:
                    ax.set_xlim(0., +xlim)
                ax.set_ylim(np.nanpercentile(self.radobj.yall,0)-np.abs(0.5*np.nanpercentile(self.radobj.yall,0)),np.nanpercentile(self.radobj.yall,99.9)+np.abs(0.25*np.nanpercentile(self.radobj.yall,99.9)))

                ax.plot(self.radobj.xall, self.radobj.yall, 'k.', markersize = 1., alpha=styles.get_scatter_alpha(len(self.radobj.xall))) 

                ##########
                if self.radobj.binning:
                    plotbinx, plotbiny = np.ravel(list(zip(self.radobj.bins[:-1], self.radobj.bins[1:]))), np.ravel(list(zip(self.radobj.mastery, self.radobj.mastery)))
                    ax.plot(plotbinx, plotbiny,
                              'r-')

                # Plot the range
                plot_bgdist = self.radobj.bgdist.copy()
                plot_bgdist[~np.isfinite(plot_bgdist)] = np.asarray(ax.get_xlim())[~np.isfinite(plot_bgdist)]
                ax.fill_between(plot_bgdist, *ax.get_ylim(),
                                  facecolor = (0., 1., 0., .05),
                                  edgecolor = 'g',
                                  linestyle = '--',
                                  linewidth = 1.)
                ax.fill_between(-plot_bgdist, *ax.get_ylim(),
                                  facecolor = (0., 1., 0., .05),
                                  edgecolor = 'g',
                                  linestyle = '--',
                                  linewidth = 1.)
                ax.plot(np.linspace(ax.get_xlim()[0], ax.get_xlim()[1], 500), self.radobj.bgfit(np.linspace(ax.get_xlim()[0], ax.get_xlim()[1], 500)),'g-', lw=3)
                #ax.set_xticklabels([])
                #ax.tick_params(labelsize=14)

                xplot = self.radobj.xall
                yplot = self.radobj.yall - self.radobj.bgfit(xplot)


                #Add labels#
                if self.radobj.bgfit.degree == 1:
                    ax.text(0.03, 0.95,"y=({:.2E})x+({:.2E})".format(self.radobj.bgfit.parameters[1],self.radobj.bgfit.parameters[0]),ha='left',va='top', fontweight='bold',fontsize=20, transform=ax.transAxes)#,bbox={'facecolor':'white', 'edgecolor':'none', 'alpha':1.0, 'pad':1})
                elif self.radobj.bgfit.degree == 0:
                    ax.text(0.03, 0.95,"y=({:.2E})".format(self.radobj.bgfit.c0.value),ha='left',va='top', fontweight='bold',fontsize=20, transform=ax.transAxes)
                else:
                    warnings.warn("Labeling BG functions of higher degrees during plotting are not supported yet.")
                ax.text(0.97, 0.95,"Background\nFit", ha='right',va='top', fontweight='bold',fontsize=20, color='green',transform=ax.transAxes)#,bbox={'facecolor':'white', 'edgecolor':'none', 'alpha':1.0, 'pad':1})

            else:
                raise ValueError('plotFeature has to be either "model" or "bg".')

        else:
            raise ValueError('plotFeature has to be either "model" or "bg".')



from matplotlib import rcParams
from cycler import cycler
import numpy as np

colors_538 = ["#30a2da",
              "#fc4f30",
              "#e5ae38",
              "#6d904f",
              "#8b8b8b"]

#Set size opacity of scatter points based on total number of sampled points 
def get_scatter_alpha(numpoints):
    if numpoints < 1000:
        return 0.9
    
    elif numpoints < 10000:
        return 0.7
    
    elif numpoints < 100000:
        return 0.5
        
    else:
        return 0.1


### Hope's implementation ###
# lines
rcParams['lines.linewidth'] = 1
rcParams['lines.markersize'] = 20
# image
rcParams['image.cmap'] = 'viridis'
rcParams['image.interpolation'] = 'none'
rcParams['image.origin'] = 'bottom'
# patch
rcParams['patch.facecolor'] = 'yellow'
rcParams['patch.edgecolor'] = 'none'
# font
rcParams['font.size'] = 26
rcParams['font.family'] = 'StixGeneral'
# legend
rcParams['legend.frameon'] = False
rcParams['legend.scatterpoints'] = 1
rcParams['legend.numpoints'] = 1
rcParams['legend.fontsize'] = 26
# axes
rcParams['axes.prop_cycle'] = cycler('color', colors_538)
rcParams['axes.facecolor'] = 'none'
# figure
rcParams['figure.figsize'] = (14, 14)
rcParams['figure.dpi'] = 180
rcParams['figure.subplot.left'] = .1
rcParams['figure.subplot.right'] = .97
rcParams['figure.subplot.bottom'] = .1
rcParams['figure.subplot.top'] = .96
rcParams['figure.subplot.wspace'] = .02
rcParams['figure.subplot.hspace'] = 0.
# savefig
rcParams['savefig.jpeg_quality'] = 100
rcParams['savefig.dpi'] = 180
rcParams['savefig.directory'] = '/Users/hopechen/Desktop/'



from distutils.core import setup

setup(
  name = 'radfil',
  packages = ['radfil'],
  version = '1.1.1',
  description = 'Build radial density profiles for interstellar filaments. See Zucker & Chen 2018 for full details. Contact catherine.zucker@cfa.harvard.edu & hopechen@utexas.edu with questions/comments!',
  author = ['Catherine Zucker', 'Hope Chen'],
  author_email = 'catherine.zucker@cfa.harvard.edu',
  url = 'http://github.com/catherinezucker/RadFil', # use the URL to the github repo
  download_url = 'https://api.github.com/repos/catherinezucker/radfil/zipball/master',
  keywords = ['astrophysics', 'radfil', 'filaments'], # arbitrary keywords
  classifiers = [],
  install_requires=[
      'numpy',
      'scipy',
      'matplotlib',
      'astropy',
      'shapely',
      'sklearn',
      'scikit-image',
      'networkx',
      'pandas'
  ]
)
